#include "rweTask.h"
#include "getFilePath.h"
#include "version.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define TASKLEN       1000
#define TEXTLEN       65535
#define MAX_TASK_NUM  62
#define FILE_PATH_LEN 512
#define CONFIG_LEN    40
#define CMD_LEN       512


enum editMode {
    EDIT_REWRITE_ALL,
    EDIT_REPLACE_FIRST,
    EDIT_REPLACE_LAST,
    EDIT_REPLACE_ALL,
    EDIT_TOPIC_TEXT
};


struct task {
    int level;
    int num;
    char *p_task;
    char *p_text;
};


int readTask(FILE *fp);
void writeTask(FILE *fp, struct task *p);
void outputTask(int outputLevel);
void outputTopic(void);
void editTaskText(char *origin, char *target, int mode);
void getColorMode(int *numColor, int *linkColor, int *textColor);
void editTopicText(char **p_text);


struct task taskOne;


void printTask(int printLevel)
{
    FILE *fp;
    char filePath[FILE_PATH_LEN];

    getFilePath(filePath);
    fp = fopen(filePath, "r");
    while (readTask(fp) == 0)
        outputTask(printLevel);
    fclose(fp);
}


void printTopic(int num)
{
    FILE *fp;
    char filePath[FILE_PATH_LEN];

    getFilePath(filePath);
    fp = fopen(filePath, "r");
    while (readTask(fp) == 0) {
        if (num == taskOne.num)
            outputTopic();
        free(taskOne.p_task);
        free(taskOne.p_text);
    }
    fclose(fp);
}


void addTask(int is_topic, int num, char *strList[])
{
    struct task taskCurr;
    char taskTemp[TASKLEN];
    FILE *fp;
    FILE *fp_swp;
    char filePath[FILE_PATH_LEN];
    char fileSwpPath[FILE_PATH_LEN];
    char numList[MAX_TASK_NUM];
    int c;
    int i = 0;

    *taskTemp = '\0';

    /* creat task number list */
    for (c = '0'; c <= '9'; c++)
        numList[i++] = c;
    for (c = 'a'; c <= 'z'; c++)
        numList[i++] = c;
    for (c = 'A'; c <= 'Z'; c++)
        numList[i++] = c;

    /* copy to swap file */
    getFilePath(filePath);
    getFileSwpPath(fileSwpPath);
    fp = fopen(filePath, "r");
    fp_swp = fopen(fileSwpPath, "w");
    i = 0;
    while (readTask(fp) == 0) {
        if (i >= MAX_TASK_NUM)
            return;
        if (numList[i++] == taskOne.num)
            writeTask(fp_swp, &taskOne);
        else {
            i--;
            taskCurr.num = (int)numList[i];
            while (num-- > 0) {
                strcat(taskTemp, *strList++);
                if (num != 0)
                    strcat(taskTemp, " ");
            }
            taskCurr.p_task = (char *)malloc(strlen(taskTemp)+1);
            taskCurr.level = 1;
            taskCurr.p_text = NULL;
            if (is_topic != 0)
                editTopicText(&taskCurr.p_text);
            strcpy(taskCurr.p_task, taskTemp);
            writeTask(fp_swp, &taskCurr);
            writeTask(fp_swp, &taskOne);
            break;
        }
    }
    while (readTask(fp) == 0)
        writeTask(fp_swp, &taskOne);
    if (num > 0) {
        taskCurr.num = (int)numList[i];
        while (num-- > 0) {
            strcat(taskTemp, *strList++);
            if (num != 0)
                strcat(taskTemp, " ");
            }
        taskCurr.p_task = (char *)malloc(strlen(taskTemp)+1);
        taskCurr.level = 1;
        taskCurr.p_text = NULL;
        if (is_topic != 0)
            editTopicText(&taskCurr.p_text);
        strcpy(taskCurr.p_task, taskTemp);
        writeTask(fp_swp, &taskCurr);
    }

    fclose(fp);
    fclose(fp_swp);
    remove(filePath);
    rename(fileSwpPath, filePath);
}


void finishTask(int num, char *strList[])
{
    FILE *fp;
    FILE *fp_swp;
    char filePath[FILE_PATH_LEN];
    char fileSwpPath[FILE_PATH_LEN];
    int i;
    int found = 0;

    getFilePath(filePath);
    getFileSwpPath(fileSwpPath);
    fp = fopen(filePath, "r");
    fp_swp = fopen(fileSwpPath, "w");

    while (readTask(fp) == 0) {
        for (i = 0; i < num; i++)
            if (taskOne.num == strList[i][0]) {
                found++;
                break;
            }
        if (i == num)
            writeTask(fp_swp, &taskOne);
    }

    fclose(fp);
    fclose(fp_swp);
    remove(filePath);
    rename(fileSwpPath, filePath);

    if (found != num) {
        printf("输入的任务序号有误\n");
        return;
    }
}


void changeTaskLevel(int taskLevel, int num, char *strList[])
{
    FILE *fp;
    FILE *fp_swp;
    char filePath[FILE_PATH_LEN];
    char fileSwpPath[FILE_PATH_LEN];
    int i;
    int found = 0;

    getFilePath(filePath);
    getFileSwpPath(fileSwpPath);
    fp = fopen(filePath, "r");
    fp_swp = fopen(fileSwpPath, "w");

    while (readTask(fp) == 0) {
        for (i = 0; i < num; i++)
            if (taskOne.num == strList[i][0]) {
                if (taskLevel >= 1 && taskLevel <= 9)
                    taskOne.level = taskLevel;
                else {
                    printf("参数格式错误，请尝试执行t --help查看帮助\n");
                    return;
                }
                found++;
                break;
            }
        writeTask(fp_swp, &taskOne);
    }

    fclose(fp);
    fclose(fp_swp);
    remove(filePath);
    rename(fileSwpPath, filePath);

    if (found != num) {
        printf("输入的任务序号有误\n");
        return;
    }
}


void editTask(int num, char *strList[])
{
    FILE *fp;
    FILE *fp_swp;
    char filePath[FILE_PATH_LEN];
    char fileSwpPath[FILE_PATH_LEN];
    int taskNum;
    char *p;
    char originStr[TASKLEN] = "";
    char targetStr[TASKLEN] = "";
    int mode;
    int c;
    int i;
    int j;

    if (num == 0)
        return;

    /* get edit mode */
    if (*(*strList+1) != '\0')
        return;
    taskNum = **strList;
    num--;
    p = *++strList;
    if (num == 0)
        mode = EDIT_TOPIC_TEXT;  /* edit only the topic text */
    else if (*p != '/' && *p != '%')
        mode = EDIT_REWRITE_ALL;  /* rewrite the task */
    else if (*p == '%')
        if (*++p == '/') {
            mode = EDIT_REPLACE_LAST;  /* replace the last one */
            p++;
        }
        else {
            printf("参数格式错误，请尝试执行t --help查看帮助\n");
            return;
        }
    else if (*p == '/')
        if (*++p == '/') {
            mode = EDIT_REPLACE_ALL;  /* replace all */
            p++;
        }
        else if (*p != '\0')
            mode = EDIT_REPLACE_FIRST;  /* replace the first one */
        else {
            printf("参数格式错误，请尝试执行t --help查看帮助\n");
            return;
        }

    /* get the whole /.../.../ */
    while (num-- > 0) {
        strcat(originStr, *strList++);
        if (num != 0)
            strcat(originStr, " ");
    }

    /* separate ... in /.../.../ */
    switch (mode) {
        case EDIT_REWRITE_ALL:
            strcpy(targetStr, originStr);
            break;
        case EDIT_REPLACE_FIRST:
            i = 1;
            break;
        case EDIT_REPLACE_LAST:
        case EDIT_REPLACE_ALL:
            i = 2;
            break;
        default:
            break;
    }
    if (mode != EDIT_REWRITE_ALL && mode != EDIT_TOPIC_TEXT) {
        j = 0;
        while ((c = originStr[i++]) != '\0' && c != '/')
            originStr[j++] = c;
        originStr[j] = '\0';
        if (c == '\0') {
            printf("参数格式错误，请尝试执行t --help查看帮助\n");
            return;
        }
        j = 0;
        while ((c = originStr[i++]) != '/' && c != '\0')
            targetStr[j++] = c;
        targetStr[j] = '\0';
    }

    /* edit task text */
    getFilePath(filePath);
    getFileSwpPath(fileSwpPath);
    fp = fopen(filePath, "r");
    fp_swp = fopen(fileSwpPath, "w");
    while (readTask(fp) == 0) {
        if (taskOne.num == taskNum) {
            editTaskText(originStr, targetStr, mode);
            if (taskOne.p_text != NULL ||
                    (mode == EDIT_TOPIC_TEXT && taskOne.p_text != NULL))
                editTopicText(&taskOne.p_text);
        }
        writeTask(fp_swp, &taskOne);
    }

    fclose(fp);
    fclose(fp_swp);
    remove(filePath);
    rename(fileSwpPath, filePath);
}


void printHelp(void)
{
    printf("\n");
    printf("these characters cannot be used:\n");
    printf("  [  ]  %%  /  \'  \"\n");
    printf("\n");
    /* add a task */
    printf("add a task:\n");
    printf("  $ t -a study Linux\n");
    printf("  $ t -a \"study math and study physics\"\n");
    printf("\n");
    /* add a topic */
    printf("add a topic:\n");
    printf("  $ t -t a study topic\n");
    printf("\n");
    /* print tasks or a topic*/
    printf("print tasks or a topic:\n");
    printf("  $ t                        # print tasks of level 1\n");
    printf(" 0 - study Linux\n");
    printf(" 1 - study math and study physics\n");
    printf(" 2 - <<  a study topic\n");
    printf("  $ t -1                     # print tasks(level 1-9)\n");
    printf("  $ t -0                     # print the whole\n");
    printf("  $ t -w                     # print the whole\n");
    printf("  $ t -T 2                   # if 2 is a topic\n");
    printf("\n");
    /* finish tasks */
    printf("finish tasks:\n");
    printf("  $ t -f 0\n");
    printf("  $ t -f 0 1\n");
    printf("\n");
    /* edit a task or a topic*/
    printf("edit a task or a topic:\n");
    printf("  $ t -e 1 learn music       # rewrite the task\n");
    printf("  $ t -e 1 /study/learn/     # replace the first 'study'\n");
    printf("  $ t -e 1 %%/study/learn/    # replace the last 'study'\n");
    printf("  $ t -e 1 //study/learn/    # replace all 'study's\n");
    printf("  $ t -e 2                   # if 2 is a topic\n");
    printf("\n");
    /* change the importance degree */
    printf("change the level:\n");
    printf("  $ t -9 0 1                 # change the level of the tasks\n");
    printf("\n");
    /* change the color of output */
    printf("change the color of output:\n");
    printf("  $ t -s\n");
    printf("\n");
}


void printVersion(void)
{
    printf("\n");
    printf("tT %s\n", VERSION);
    printf("Copyright (C) %s %s\n", YEAR, AUTHOR1);
    printf("License GPLv%s+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n", GPLVER);
    printf("这是自由软件：您可以自由地更改并重新发布它。\n");
    printf("在法律范围内没有其它保证。\n");
    printf("\n");
    printf("by %s <%s>\n", AUTHOR1, AUTHOR1_EMAIL);
    printf("\n");
}


int readTask(FILE *fp)
{
    int c;
    char taskTemp[TASKLEN];
    char textTemp[TEXTLEN];
    int p_taskTemp = 0;
    int p_textTemp = 0;

    /* read task number */
    if ((c = fgetc(fp)) == EOF)
        return 1;
    taskOne.num = c;

    /* read task level */
    taskOne.level = fgetc(fp) - '0';

    /* read task description */
    while ((c = fgetc(fp)) != '[')
        taskTemp[p_taskTemp++] = c;
    taskTemp[p_taskTemp] = '\0';
    taskOne.p_task = (char *)malloc(strlen(taskTemp)+1);
    if (taskOne.p_task != NULL)
        strcpy(taskOne.p_task, taskTemp);

    /* read task text */
    while ((c = fgetc(fp)) != '[')
        textTemp[p_textTemp++] = c;
    textTemp[p_textTemp] = '\0';
    if (p_textTemp == 0)
        taskOne.p_text = NULL;
    else {
        taskOne.p_text = (char *)malloc(strlen(textTemp)+1);
        if (taskOne.p_text != NULL)
            strcpy(taskOne.p_text, textTemp);
    }
    fgetc(fp);

    return 0;
}


void writeTask(FILE *fp, struct task *p)
{
    fputc(p->num, fp);
    fputc(p->level + '0', fp);
    fprintf(fp, "%s", p->p_task);
    fputc('[', fp);
    if (p->p_text != NULL)
        fprintf(fp, "%s", p->p_text);
    fputc('[', fp);
    fputc('\n', fp);

    free(p->p_task);
    free(p->p_text);
}


void outputTask(int outputLevel)
{
    int numColor;
    int linkColor;
    int textColor;

    if (outputLevel != 0 && outputLevel != taskOne.level) {
        free(taskOne.p_task);
        return;
    }

    getColorMode(&numColor, &linkColor, &textColor);
    printf(" ");
    /* print number */
    if (numColor == 0)
        printf("%c", taskOne.num);
    else
        printf("\033[01;%dm%c\033[00m", numColor, taskOne.num);
    /* print - */
    if (linkColor == 0)
        printf(" - ");
    else
        printf(" \033[01;%dm-\033[00m ", linkColor);
    /* print << if this task is a topic */
    if (taskOne.p_text != NULL)
        printf("<<  ");
    /* print text */
    if (textColor == 0)
        printf("%s", taskOne.p_task);
    else
        printf("\033[01;%dm%s\033[00m", textColor, taskOne.p_task);
    putchar('\n');

    free(taskOne.p_task);
    free(taskOne.p_text);
}


void outputTopic(void)
{
    int numColor;
    int linkColor;
    int textColor;

    getColorMode(&numColor, &linkColor, &textColor);
    printf(" ");
    if (taskOne.p_text != NULL)
        printf("   >>  ");
    /* print number */
    if (numColor == 0)
        printf("%c", taskOne.num);
    else
        printf("\033[01;%dm%c\033[00m", numColor, taskOne.num);
    /* print - */
    if (linkColor == 0)
        printf(" - ");
    else
        printf(" \033[01;%dm-\033[00m ", linkColor);
    /* print text */
    if (textColor == 0)
        printf("%s", taskOne.p_task);
    else
        printf("\033[01;%dm%s\033[00m", textColor, taskOne.p_task);
    putchar('\n');
    /* print topic text */
    if (taskOne.p_text != NULL && strcmp(taskOne.p_text, "]") != 0)
        printf("%s", taskOne.p_text);
}


void editTaskText(char *origin, char *target, int mode)
{
    char text[TASKLEN];
    char tempText[TASKLEN];
    char *p;
    char *q;
    char *r;

    strcpy(tempText, taskOne.p_task);
    free(taskOne.p_task);

    if (mode == EDIT_REWRITE_ALL)
        strcpy(text, target);
    else if (mode == EDIT_REPLACE_FIRST) {
        strcpy(text, tempText);
        if ((p = strstr(tempText, origin)) != NULL) {
            strcpy(text+(p-tempText), target);
            strcpy(text+(p-tempText)+strlen(target), p+strlen(origin));
        }
    }
    else if (mode == EDIT_REPLACE_LAST) {
        p = tempText;
        while ((q = strstr(p, origin)) != NULL) {
            r = q;
            p = r + 1;
        }
        strcpy(text, tempText);
        if (r != NULL) {
            strcpy(text+(r-tempText), target);
            strcpy(text+(r-tempText)+strlen(target), r+strlen(origin));
        }
    }
    else if (mode == EDIT_REPLACE_ALL) {
        strcpy(text, tempText);
        p = tempText;
        q = text;
        while ((r = strstr(p, origin)) != NULL) {
            strcpy(q+(r-p), target);
            q = q + (r-p) + strlen(target);
            p = r + strlen(origin);
            strcpy(q, p);
        }
    }

    taskOne.p_task = (char *)malloc(strlen(text)+1);
    strcpy(taskOne.p_task, text);
}


void editTopicText(char **p_text)
{
    FILE *fp;
    char textPath[FILE_PATH_LEN];
    char cmd[CMD_LEN];
    char text[TEXTLEN];
    char *p;
    int c;

    getTextFilePath(textPath);
    sprintf(cmd, "%s%s", "touch ", textPath);
    system(cmd);
    sprintf(cmd, "%s%s", "vi ", textPath);

    if (*p_text == NULL) {
        system(cmd);
        p = text;
        fp = fopen(textPath, "r");
        while ((c = fgetc(fp)) != EOF)
            *p++ = c;
        if (p == text)
            *p++ = ']';
        *p = '\0';
        fclose(fp);
        remove(textPath);
        *p_text = (char *)malloc(strlen(text)+1);
        strcpy(*p_text, text);
    }
    else {
        p = *p_text;
        fp = fopen(textPath, "w");
        while (*p != '\0') {
            if (*p != ']')
                fputc(*p, fp);
            ++p;
        }
        fclose(fp);
        system(cmd);
        p = text;
        fp = fopen(textPath, "r");
        while ((c = fgetc(fp)) != EOF)
            *p++ = c;
        if (p == text)
            *p++ = ']';
        *p = '\0';
        fclose(fp);
        remove(textPath);
        *p_text = (char *)malloc(strlen(text)+1);
        strcpy(*p_text, text);
    }
}


void getColorMode(int *numColor, int *linkColor, int *textColor)
{
    FILE *fp;
    char configPath[FILE_PATH_LEN];
    char line[CONFIG_LEN];
    char colorMode;

    getConfigPath(configPath);
    fp = fopen(configPath, "r");
    colorMode = '0';
    while (fgets(line, CONFIG_LEN, fp) != NULL)
        if (strncmp(line, "set color ", 10) == 0)
            colorMode = line[10];
    fclose(fp);

    if (colorMode == '0')
        *numColor = *linkColor = *textColor = 0;
    else if (colorMode >= '1' && colorMode <= '6') {
        *numColor = *textColor = 30 + (colorMode - '0');
        *linkColor = 0;
    }
    else if (colorMode == 'a') {
        *numColor = *textColor = 0;
        *linkColor = 33;
    }
    else if (colorMode >= 'b' && colorMode <= 'g') {
        *numColor = 30 + (colorMode - 'a');
        *linkColor = *numColor == 34 ? 33 : 34;
        *textColor = 0;
    }
    else {
        switch(colorMode) {
            case 'h':
                *numColor = 32;
                *linkColor = 34;
                *textColor = 33;
                break;
            case 'i':
                *numColor = 33;
                *linkColor = 35;
                *textColor = 32;
                break;
            case 'j':
                *numColor = 36;
                *linkColor = 35;
                *textColor = 33;
                break;
            case 'k':
                *numColor = 33;
                *linkColor = 34;
                *textColor = 36;
                break;
            case 'l':
                *numColor = 32;
                *linkColor = 35;
                *textColor = 36;
                break;
            case 'm':
                *numColor = 36;
                *linkColor = 34;
                *textColor = 32;
                break;
            default:
                printf("Config file error\n");
                break;
        }
    }
}
